@@ -9,10 +9,8 @@ import (
 	"sigs.k8s.io/yaml"
 
 	"github.com/gophercloud/gophercloud"
-	"github.com/gophercloud/gophercloud/openstack/baremetal/noauth"
 	"github.com/gophercloud/gophercloud/openstack/baremetal/v1/nodes"
 	"github.com/gophercloud/gophercloud/openstack/baremetal/v1/ports"
-	noauthintrospection "github.com/gophercloud/gophercloud/openstack/baremetalintrospection/noauth"
 	"github.com/gophercloud/gophercloud/openstack/baremetalintrospection/v1/introspection"
 
 	"github.com/pkg/errors"
@@ -24,6 +22,8 @@ import (
 	"github.com/metal3-io/baremetal-operator/pkg/bmc"
 	"github.com/metal3-io/baremetal-operator/pkg/hardware"
 	"github.com/metal3-io/baremetal-operator/pkg/provisioner"
+	"github.com/metal3-io/baremetal-operator/pkg/provisioner/ironic/clients"
+	"github.com/metal3-io/baremetal-operator/pkg/provisioner/ironic/devicehints"
 	"github.com/metal3-io/baremetal-operator/pkg/provisioner/ironic/hardwaredetails"
 )
 
@@ -32,21 +32,38 @@ var deprovisionRequeueDelay = time.Second * 10
 var provisionRequeueDelay = time.Second * 10
 var powerRequeueDelay = time.Second * 10
 var introspectionRequeueDelay = time.Second * 15
+var softPowerOffTimeout = time.Second * 180
 var deployKernelURL string
 var deployRamdiskURL string
 var ironicEndpoint string
 var inspectorEndpoint string
+var ironicAuth clients.AuthConfig
+var inspectorAuth clients.AuthConfig
 
 const (
 	// See nodes.Node.PowerState for details
-	powerOn   = "power on"
-	powerOff  = "power off"
-	powerNone = "None"
+	powerOn      = "power on"
+	powerOff     = "power off"
+	softPowerOff = "soft power off"
+	powerNone    = "None"
 )
 
+var bootModeCapabilities = map[metal3v1alpha1.BootMode]string{
+	metal3v1alpha1.UEFI:   "boot_mode:uefi",
+	metal3v1alpha1.Legacy: "boot_mode:bios",
+}
+
 func init() {
 	// NOTE(dhellmann): Use Fprintf() to report errors instead of
 	// logging, because logging is not configured yet in init().
+
+	var authErr error
+	ironicAuth, inspectorAuth, authErr = clients.LoadAuth()
+	if authErr != nil {
+		fmt.Fprintf(os.Stderr, "Cannot start: %s\n", authErr)
+		os.Exit(1)
+	}
+
 	deployKernelURL = os.Getenv("DEPLOY_KERNEL_URL")
 	if deployKernelURL == "" {
 		fmt.Fprintf(os.Stderr, "Cannot start: No DEPLOY_KERNEL_URL variable set\n")
@@ -95,7 +112,9 @@ type ironicProvisioner struct {
 func LogStartup() {
 	log.Info("ironic settings",
 		"endpoint", ironicEndpoint,
+		"ironicAuthType", ironicAuth.Type,
 		"inspectorEndpoint", inspectorEndpoint,
+		"inspectorAuthType", inspectorAuth.Type,
 		"deployKernelURL", deployKernelURL,
 		"deployRamdiskURL", deployRamdiskURL,
 	)
@@ -104,36 +123,35 @@ func LogStartup() {
 // A private function to construct an ironicProvisioner (rather than a
 // Provisioner interface) in a consistent way for tests.
 func newProvisioner(host *metal3v1alpha1.BareMetalHost, bmcCreds bmc.Credentials, publisher provisioner.EventPublisher) (*ironicProvisioner, error) {
-	client, err := noauth.NewBareMetalNoAuth(noauth.EndpointOpts{
-		IronicEndpoint: ironicEndpoint,
-	})
+	clientIronic, err := clients.IronicClient(ironicEndpoint, ironicAuth)
 	if err != nil {
 		return nil, err
 	}
+
+	clientInspector, err := clients.InspectorClient(inspectorEndpoint, inspectorAuth)
+	if err != nil {
+		return nil, err
+	}
+
 	bmcAccess, err := bmc.NewAccessDetails(host.Spec.BMC.Address, host.Spec.BMC.DisableCertificateVerification)
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to parse BMC address information")
 	}
-	inspector, err := noauthintrospection.NewBareMetalIntrospectionNoAuth(
-		noauthintrospection.EndpointOpts{
-			IronicInspectorEndpoint: inspectorEndpoint,
-		})
-	if err != nil {
-		return nil, err
-	}
+
 	// Ensure we have a microversion high enough to get the features
 	// we need.
-	client.Microversion = "1.56"
+	clientIronic.Microversion = "1.56"
 	p := &ironicProvisioner{
 		host:      host,
 		status:    &(host.Status.Provisioning),
 		bmcAccess: bmcAccess,
 		bmcCreds:  bmcCreds,
-		client:    client,
-		inspector: inspector,
+		client:    clientIronic,
+		inspector: clientInspector,
 		log:       log.WithValues("host", host.Name),
 		publisher: publisher,
 	}
+
 	return p, nil
 }
 
@@ -222,7 +240,6 @@ func (p *ironicProvisioner) ValidateManagementAccess(credentialsChanged bool) (r
 		msg := fmt.Sprintf("BMC driver %s requires a BootMACAddress value", p.bmcAccess.Type())
 		p.log.Info(msg)
 		result.ErrorMessage = msg
-		result.Dirty = true
 		return result, nil
 	}
 
@@ -248,6 +265,9 @@ func (p *ironicProvisioner) ValidateManagementAccess(credentialsChanged bool) (r
 				PowerInterface:      p.bmcAccess.PowerInterface(),
 				RAIDInterface:       p.bmcAccess.RAIDInterface(),
 				VendorInterface:     p.bmcAccess.VendorInterface(),
+				Properties: map[string]interface{}{
+					"capabilities": bootModeCapabilities[p.host.Status.Provisioning.BootMode],
+				},
 			}).Extract()
 		// FIXME(dhellmann): Handle 409 and 503? errors here.
 		if err != nil {
@@ -279,12 +299,13 @@ func (p *ironicProvisioner) ValidateManagementAccess(credentialsChanged bool) (r
 			}
 		}
 
-		if p.host.Spec.Image != nil && p.host.Spec.Image.URL != "" {
-			checksum := p.host.Spec.Image.Checksum
+		checksum, checksumType, ok := p.host.GetImageChecksum()
 
+		if ok {
 			p.log.Info("setting instance info",
 				"image_source", p.host.Spec.Image.URL,
-				"checksum", checksum,
+				"image_os_hash_value", checksum,
+				"image_os_hash_algo", checksumType,
 			)
 
 			updates := nodes.UpdateOpts{
@@ -295,15 +316,48 @@ func (p *ironicProvisioner) ValidateManagementAccess(credentialsChanged bool) (r
 				},
 				nodes.UpdateOperation{
 					Op:    nodes.AddOp,
-					Path:  "/instance_info/image_checksum",
+					Path:  "/instance_info/image_os_hash_value",
 					Value: checksum,
 				},
+				nodes.UpdateOperation{
+					Op:    nodes.AddOp,
+					Path:  "/instance_info/image_os_hash_algo",
+					Value: checksumType,
+				},
 				nodes.UpdateOperation{
 					Op:    nodes.ReplaceOp,
 					Path:  "/instance_uuid",
 					Value: string(p.host.ObjectMeta.UID),
 				},
 			}
+
+			// image_checksum
+			//
+			// FIXME: For older versions of ironic that do not have
+			// https://review.opendev.org/#/c/711816/ failing to
+			// include the 'image_checksum' causes ironic to refuse to
+			// provision the image, even if the other hash value
+			// parameters are given. We only want to do that for MD5,
+			// however, because those versions of ironic only support
+			// MD5 checksums.
+			if checksumType == string(metal3v1alpha1.MD5) {
+				updates = append(
+					updates,
+					nodes.UpdateOperation{
+						Op:    nodes.AddOp,
+						Path:  "/instance_info/image_checksum",
+						Value: checksum,
+					},
+				)
+			}
+
+			if p.host.Spec.Image.DiskFormat != nil {
+				updates = append(updates, nodes.UpdateOperation{
+					Op:    nodes.AddOp,
+					Path:  "/instance_info/image_disk_format",
+					Value: *p.host.Spec.Image.DiskFormat,
+				})
+			}
 			_, err = nodes.Update(p.client, ironicNode.UUID, updates).Extract()
 			switch err.(type) {
 			case nil:
@@ -411,12 +465,13 @@ func (p *ironicProvisioner) ValidateManagementAccess(credentialsChanged bool) (r
 		// If we're still waiting for the state to change in Ironic,
 		// return true to indicate that we're dirty and need to be
 		// reconciled again.
+		result.RequeueAfter = provisionRequeueDelay
 		result.Dirty = true
 		return result, nil
 	}
 }
 
-func (p *ironicProvisioner) changeNodeProvisionState(ironicNode *nodes.Node, opts nodes.ProvisionStateOpts) (result provisioner.Result, err error) {
+func (p *ironicProvisioner) tryChangeNodeProvisionState(ironicNode *nodes.Node, opts nodes.ProvisionStateOpts) (success bool, result provisioner.Result, err error) {
 	p.log.Info("changing provisioning state",
 		"current", ironicNode.ProvisionState,
 		"existing target", ironicNode.TargetProvisionState,
@@ -426,16 +481,23 @@ func (p *ironicProvisioner) changeNodeProvisionState(ironicNode *nodes.Node, opt
 	changeResult := nodes.ChangeProvisionState(p.client, ironicNode.UUID, opts)
 	switch changeResult.Err.(type) {
 	case nil:
+		success = true
 	case gophercloud.ErrDefault409:
 		p.log.Info("could not change state of host, busy")
 	default:
-		return result, errors.Wrap(changeResult.Err,
+		err = errors.Wrap(changeResult.Err,
 			fmt.Sprintf("failed to change provisioning state to %q", opts.Target))
+		return
 	}
 
 	result.Dirty = true
 	result.RequeueAfter = provisionRequeueDelay
-	return result, nil
+	return
+}
+
+func (p *ironicProvisioner) changeNodeProvisionState(ironicNode *nodes.Node, opts nodes.ProvisionStateOpts) (result provisioner.Result, err error) {
+	_, result, err = p.tryChangeNodeProvisionState(ironicNode, opts)
+	return
 }
 
 // InspectHardware updates the HardwareDetails field of the host with
@@ -465,12 +527,34 @@ func (p *ironicProvisioner) InspectHardware() (result provisioner.Result, detail
 				err = nil
 				return
 			default:
+				p.log.Info("updating boot mode before hardware inspection")
+				op, value := buildCapabilitiesValue(ironicNode, p.host.Status.Provisioning.BootMode)
+				updates := nodes.UpdateOpts{
+					nodes.UpdateOperation{
+						Op:    op,
+						Path:  "/properties/capabilities",
+						Value: value,
+					},
+				}
+				_, err = nodes.Update(p.client, ironicNode.UUID, updates).Extract()
+				switch err.(type) {
+				case nil:
+				case gophercloud.ErrDefault409:
+					p.log.Info("could not update host settings in ironic, busy")
+					result.Dirty = true
+					return
+				default:
+					err = errors.Wrap(err, "failed to update host boot mode settings in ironic")
+					return
+				}
+
 				p.log.Info("starting new hardware inspection")
-				result, err = p.changeNodeProvisionState(
+				var success bool
+				success, result, err = p.tryChangeNodeProvisionState(
 					ironicNode,
 					nodes.ProvisionStateOpts{Target: nodes.TargetInspect},
 				)
-				if err == nil {
+				if success {
 					p.publisher("InspectionStarted", "Hardware inspection started")
 				}
 				return
@@ -544,9 +628,10 @@ func (p *ironicProvisioner) UpdateHardwareState() (result provisioner.Result, er
 	return result, nil
 }
 
-func (p *ironicProvisioner) getUpdateOptsForNode(ironicNode *nodes.Node, checksum string) (updates nodes.UpdateOpts, err error) {
+func (p *ironicProvisioner) getUpdateOptsForNode(ironicNode *nodes.Node) (updates nodes.UpdateOpts, err error) {
 
 	hwProf, err := hardware.GetProfile(p.host.HardwareProfile())
+
 	if err != nil {
 		return updates, errors.Wrap(err,
 			fmt.Sprintf("Could not start provisioning with bad hardware profile %s",
@@ -571,23 +656,76 @@ func (p *ironicProvisioner) getUpdateOptsForNode(ironicNode *nodes.Node, checksu
 		},
 	)
 
-	// image_checksum
-	if _, ok := ironicNode.InstanceInfo["image_checksum"]; !ok {
+	checksum, checksumType, _ := p.host.GetImageChecksum()
+
+	// image_os_hash_algo
+	if _, ok := ironicNode.InstanceInfo["image_os_hash_algo"]; !ok {
 		op = nodes.AddOp
-		p.log.Info("adding image_checksum")
+		p.log.Info("adding image_os_hash_algo")
 	} else {
 		op = nodes.ReplaceOp
-		p.log.Info("updating image_checksum")
+		p.log.Info("updating image_os_hash_algo")
 	}
 	updates = append(
 		updates,
 		nodes.UpdateOperation{
 			Op:    op,
-			Path:  "/instance_info/image_checksum",
+			Path:  "/instance_info/image_os_hash_algo",
+			Value: checksumType,
+		},
+	)
+
+	// image_os_hash_value
+	if _, ok := ironicNode.InstanceInfo["image_os_hash_value"]; !ok {
+		op = nodes.AddOp
+		p.log.Info("adding image_os_hash_value")
+	} else {
+		op = nodes.ReplaceOp
+		p.log.Info("updating image_os_hash_value")
+	}
+	updates = append(
+		updates,
+		nodes.UpdateOperation{
+			Op:    op,
+			Path:  "/instance_info/image_os_hash_value",
 			Value: checksum,
 		},
 	)
 
+	// image_checksum
+	//
+	// FIXME: For older versions of ironic that do not have
+	// https://review.opendev.org/#/c/711816/ failing to include the
+	// 'image_checksum' causes ironic to refuse to provision the
+	// image, even if the other hash value parameters are given. We
+	// only want to do that for MD5, however, because those versions
+	// of ironic only support MD5 checksums.
+	if checksumType == string(metal3v1alpha1.MD5) {
+		if _, ok := ironicNode.InstanceInfo["image_checksum"]; !ok {
+			op = nodes.AddOp
+			p.log.Info("adding image_checksum")
+		} else {
+			op = nodes.ReplaceOp
+			p.log.Info("updating image_checksum")
+		}
+		updates = append(
+			updates,
+			nodes.UpdateOperation{
+				Op:    op,
+				Path:  "/instance_info/image_checksum",
+				Value: checksum,
+			},
+		)
+	}
+
+	if p.host.Spec.Image.DiskFormat != nil {
+		updates = append(updates, nodes.UpdateOperation{
+			Op:    nodes.AddOp,
+			Path:  "/instance_info/image_disk_format",
+			Value: *p.host.Spec.Image.DiskFormat,
+		})
+	}
+
 	// instance_uuid
 	p.log.Info("setting instance_uuid")
 	updates = append(
@@ -633,13 +771,12 @@ func (p *ironicProvisioner) getUpdateOptsForNode(ironicNode *nodes.Node, checksu
 		op = nodes.ReplaceOp
 		p.log.Info("updating root_device")
 	}
-	hints := map[string]string{}
-	switch {
-	case hwProf.RootDeviceHints.DeviceName != "":
-		hints["name"] = hwProf.RootDeviceHints.DeviceName
-	case hwProf.RootDeviceHints.HCTL != "":
-		hints["hctl"] = hwProf.RootDeviceHints.HCTL
-	}
+
+	// hints
+	//
+	// If the user has provided explicit root device hints, they take
+	// precedence. Otherwise use the values from the hardware profile.
+	hints := devicehints.MakeHintMap(p.host.Status.Provisioning.RootDeviceHints)
 	p.log.Info("using root device", "hints", hints)
 	updates = append(
 		updates,
@@ -687,14 +824,74 @@ func (p *ironicProvisioner) getUpdateOptsForNode(ironicNode *nodes.Node, checksu
 		},
 	)
 
+	// boot_mode
+	op, value := buildCapabilitiesValue(ironicNode, p.host.Status.Provisioning.BootMode)
+	updates = append(
+		updates,
+		nodes.UpdateOperation{
+			Op:    op,
+			Path:  "/properties/capabilities",
+			Value: value,
+		},
+	)
+
 	return updates, nil
 }
 
-func (p *ironicProvisioner) startProvisioning(ironicNode *nodes.Node, checksum string, hostConf provisioner.HostConfigData) (result provisioner.Result, err error) {
+// We can't just replace the capabilities because we need to keep the
+// values provided by inspection. We can't replace only the boot_mode
+// because the API isn't fine-grained enough for that. So we have to
+// look at the existing value and modify it. This function
+// encapsulates the logic for building the value and knowing which
+// update operation to use with the results.
+func buildCapabilitiesValue(ironicNode *nodes.Node, bootMode metal3v1alpha1.BootMode) (op nodes.UpdateOp, value string) {
 
-	p.log.Info("starting provisioning")
+	capabilities, ok := ironicNode.Properties["capabilities"]
+	if !ok {
+		// There is no existing capabilities value
+		return nodes.AddOp, bootModeCapabilities[bootMode]
+	}
+	existingCapabilities := capabilities.(string)
 
-	updates, err := p.getUpdateOptsForNode(ironicNode, checksum)
+	// The capabilities value is set, so we will want to replace it.
+	op = nodes.ReplaceOp
+
+	if existingCapabilities == "" {
+		// The existing value is empty so we can replace the whole
+		// thing.
+		value = bootModeCapabilities[bootMode]
+		return
+	}
+
+	if !strings.Contains(existingCapabilities, "boot_mode") {
+		// No boot_mode is set, append and return
+		value = fmt.Sprintf("%s,%s", existingCapabilities, bootModeCapabilities[bootMode])
+		return
+	}
+
+	// The capabilities value has format "var1:val1,var2:val2". We
+	// know that boot_mode is there but not what value it has.  There
+	// are only 2 values, so we can replace the "wrong" string with
+	// the right one without fully parsing the string. We may want to
+	// change this later when we have more boot modes.
+
+	var fromMode metal3v1alpha1.BootMode
+	if bootMode == metal3v1alpha1.UEFI {
+		fromMode = metal3v1alpha1.Legacy
+	} else {
+		fromMode = metal3v1alpha1.UEFI
+	}
+	value = strings.ReplaceAll(existingCapabilities,
+		bootModeCapabilities[fromMode], bootModeCapabilities[bootMode])
+
+	return
+}
+
+func (p *ironicProvisioner) startProvisioning(ironicNode *nodes.Node, hostConf provisioner.HostConfigData) (result provisioner.Result, err error) {
+
+	p.log.Info("starting provisioning", "node properties", ironicNode.Properties)
+
+	updates, err := p.getUpdateOptsForNode(ironicNode)
 	_, err = nodes.Update(p.client, ironicNode.UUID, updates).Extract()
 	switch err.(type) {
 	case nil:
@@ -801,16 +998,18 @@ func (p *ironicProvisioner) Provision(hostConf provisioner.HostConfigData) (resu
 
 	p.log.Info("provisioning image to host", "state", ironicNode.ProvisionState)
 
-	checksum := p.host.Spec.Image.Checksum
+	checksum, checksumType, _ := p.host.GetImageChecksum()
 
 	// Local variable to make it easier to test if ironic is
 	// configured with the same image we are trying to provision to
 	// the host.
 	ironicHasSameImage := (ironicNode.InstanceInfo["image_source"] == p.host.Spec.Image.URL &&
-		ironicNode.InstanceInfo["image_checksum"] == checksum)
+		ironicNode.InstanceInfo["image_os_hash_algo"] == checksumType &&
+		ironicNode.InstanceInfo["image_os_hash_value"] == checksum)
 	p.log.Info("checking image settings",
 		"source", ironicNode.InstanceInfo["image_source"],
-		"checksum", checksum,
+		"image_os_hash_algo", checksumType,
+		"image_os_has_value", checksum,
 		"same", ironicHasSameImage,
 		"provisionState", ironicNode.ProvisionState)
 
@@ -839,42 +1038,55 @@ func (p *ironicProvisioner) Provision(hostConf provisioner.HostConfigData) (resu
 			return result, nil
 		}
 		p.log.Info("recovering from previous failure")
-		return p.startProvisioning(ironicNode, checksum, hostConf)
+		return p.startProvisioning(ironicNode, hostConf)
 
 	case nodes.Manageable:
-		return p.startProvisioning(ironicNode, checksum, hostConf)
+		return p.startProvisioning(ironicNode, hostConf)
 
 	case nodes.Available:
 		// After it is available, we need to start provisioning by
 		// setting the state to "active".
 		p.log.Info("making host active")
 
+		// Retrieve cloud-init user data
 		userData, err := hostConf.UserData()
 		if err != nil {
 			return result, errors.Wrap(err, "could not retrieve user data")
 		}
 
+		// Retrieve cloud-init network_data.json. Default value is empty
 		networkDataRaw, err := hostConf.NetworkData()
-
+		if err != nil {
+			return result, errors.Wrap(err, "could not retrieve network data")
+		}
 		var networkData map[string]interface{}
 		if err = yaml.Unmarshal([]byte(networkDataRaw), &networkData); err != nil {
 			return result, errors.Wrap(err, "failed to unmarshal network_data.json from secret")
 		}
 
+		// Retrieve cloud-init meta_data.json with falback to default
+		metaData := map[string]interface{}{
+			"uuid":             string(p.host.ObjectMeta.UID),
+			"metal3-namespace": p.host.ObjectMeta.Namespace,
+			"metal3-name":      p.host.ObjectMeta.Name,
+			"local-hostname":   p.host.ObjectMeta.Name,
+			"local_hostname":   p.host.ObjectMeta.Name,
+		}
+		metaDataRaw, err := hostConf.MetaData()
+		if err != nil {
+			return result, errors.Wrap(err, "could not retrieve metadata")
+		}
+		if metaDataRaw != "" {
+			if err = yaml.Unmarshal([]byte(metaDataRaw), &metaData); err != nil {
+				return result, errors.Wrap(err, "failed to unmarshal metadata from secret")
+			}
+		}
+
 		var configDrive nodes.ConfigDrive
 		if userData != "" {
 			configDrive = nodes.ConfigDrive{
-				UserData: userData,
-				// cloud-init requires that meta_data.json exists and
-				// that the "uuid" field is present to process
-				// any of the config drive contents.
-				MetaData: map[string]interface{}{
-					"uuid":             string(p.host.ObjectMeta.UID),
-					"metal3-namespace": p.host.ObjectMeta.Namespace,
-					"metal3-name":      p.host.ObjectMeta.Name,
-					"local-hostname":   p.host.ObjectMeta.Name,
-					"local_hostname":   p.host.ObjectMeta.Name,
-				},
+				UserData:    userData,
+				MetaData:    metaData,
 				NetworkData: networkData,
 			}
 			if err != nil {
@@ -1104,12 +1316,17 @@ func (p *ironicProvisioner) changePower(ironicNode *nodes.Node, target nodes.Tar
 		return result, nil
 	}
 
+	powerStateOpts := nodes.PowerStateOpts{
+		Target: target,
+	}
+	if target == softPowerOff {
+		powerStateOpts.Timeout = int(softPowerOffTimeout.Seconds())
+	}
+
 	changeResult := nodes.ChangePowerState(
 		p.client,
 		ironicNode.UUID,
-		nodes.PowerStateOpts{
-			Target: target,
-		})
+		powerStateOpts)
 
 	switch changeResult.Err.(type) {
 	case nil:
@@ -1119,7 +1336,11 @@ func (p *ironicProvisioner) changePower(ironicNode *nodes.Node, target nodes.Tar
 		p.log.Info("host is locked, trying again after delay", "delay", powerRequeueDelay)
 		result.Dirty = true
 		result.RequeueAfter = powerRequeueDelay
-		return result, nil
+		return result, HostLockedError{Address: p.host.Spec.BMC.Address}
+	case gophercloud.ErrDefault400:
+		// Error 400 Bad Request means target power state is not supported by vendor driver
+		p.log.Info("power change error", "message", changeResult.Err)
+		return result, SoftPowerOffUnsupportedError{Address: p.host.Spec.BMC.Address}
 	default:
 		p.log.Info("power change error", "message", changeResult.Err)
 		return result, errors.Wrap(changeResult.Err, "failed to change power state")
@@ -1151,7 +1372,6 @@ func (p *ironicProvisioner) PowerOn() (result provisioner.Result, err error) {
 		}
 		result, err = p.changePower(ironicNode, nodes.PowerOn)
 		if err != nil {
-			result.RequeueAfter = powerRequeueDelay
 			return result, errors.Wrap(err, "failed to power on host")
 		}
 		p.publisher("PowerOn", "Host powered on")
@@ -1165,6 +1385,29 @@ func (p *ironicProvisioner) PowerOn() (result provisioner.Result, err error) {
 func (p *ironicProvisioner) PowerOff() (result provisioner.Result, err error) {
 	p.log.Info("ensuring host is powered off")
 
+	result, err = p.softPowerOff()
+	if err != nil {
+		switch err.(type) {
+		// In case of soft power off is unsupported or has failed,
+		// we activate hard power off.
+		case SoftPowerOffUnsupportedError, SoftPowerOffFailed:
+			return p.hardPowerOff()
+		case HostLockedError:
+			result.RequeueAfter = powerRequeueDelay
+			result.Dirty = true
+			return result, nil
+		default:
+			result.RequeueAfter = powerRequeueDelay
+			return result, err
+		}
+	}
+	return result, nil
+}
+
+// hardPowerOff sends 'power off' request to BM node and waits for the result
+func (p *ironicProvisioner) hardPowerOff() (result provisioner.Result, err error) {
+	p.log.Info("ensuring host is powered off by \"hard power off\" command")
+
 	ironicNode, err := p.findExistingHost()
 	if err != nil {
 		return result, errors.Wrap(err, "failed to find existing host")
@@ -1179,7 +1422,6 @@ func (p *ironicProvisioner) PowerOff() (result provisioner.Result, err error) {
 		}
 		result, err = p.changePower(ironicNode, nodes.PowerOff)
 		if err != nil {
-			result.RequeueAfter = powerRequeueDelay
 			return result, errors.Wrap(err, "failed to power off host")
 		}
 		p.publisher("PowerOff", "Host powered off")
@@ -1187,3 +1429,49 @@ func (p *ironicProvisioner) PowerOff() (result provisioner.Result, err error) {
 
 	return result, nil
 }
+
+// softPowerOff sends 'soft power off' request to BM node.
+// If soft power off is not supported, the request ends with an error.
+// Otherwise the request ends with no error and the result should be
+// checked later via node fields "power_state", "target_power_state"
+// and "last_error".
+func (p *ironicProvisioner) softPowerOff() (result provisioner.Result, err error) {
+	p.log.Info("ensuring host is powered off by \"soft power off\" command")
+
+	ironicNode, err := p.findExistingHost()
+	if err != nil {
+		return result, errors.Wrap(err, "failed to find existing host")
+	}
+
+	if ironicNode.PowerState != powerOff {
+		targetState := ironicNode.TargetPowerState
+		// If the target state is either powerOff or softPowerOff, then we should wait
+		if targetState == powerOff || targetState == softPowerOff {
+			p.log.Info("waiting for power status to change")
+			result.RequeueAfter = powerRequeueDelay
+			result.Dirty = true
+			return result, nil
+		}
+		// If the target state is unset while the last error is set,
+		// then the last execution of soft power off has failed.
+		if targetState == "" && ironicNode.LastError != "" {
+			return result, SoftPowerOffFailed{Address: p.host.Spec.BMC.Address}
+		}
+		result, err = p.changePower(ironicNode, nodes.SoftPowerOff)
+		if err != nil {
+			result.RequeueAfter = powerRequeueDelay
+			return result, err
+		}
+		p.publisher("PowerOff", "Host soft powered off")
+	}
+
+	return result, nil
+}
+
+// IsReady checks if the provisioning backend is available
+func (p *ironicProvisioner) IsReady() (result bool, err error) {
+	p.log.Info("verifying ironic provisioner dependencies")
+
+	checker := newIronicDependenciesChecker(p.client, p.inspector, p.log)
+	return checker.IsReady()
+}
